# __init__.py

import fastrlapi.agents
import fastrlapi.callbacks.duration
import fastrlapi.callbacks.log
import fastrlapi.callbacks.plot
import fastrlapi.callbacks.save
import fastrlapi.core
import fastrlapi.env

# Version of the easy_agents package
__version__ = "1.5.0"


